#include "globals.h"

#define RC_FILE ".mfmacc"

void createWidgets( 
#if NeedFunctionPrototypes
	Widget, struct Accelerator * 
#endif
);
void doneCB( 
#if NeedFunctionPrototypes
	Widget, XtPointer, XtPointer 
#endif
);
void deleteCB( 
#if NeedFunctionPrototypes
	Widget, XtPointer, XtPointer 
#endif
);
void accelCB( 
#if NeedFunctionPrototypes
	Widget, XtPointer, XtPointer 
#endif
);
void singleClick( 
#if NeedFunctionPrototypes
	XtPointer 
#endif
);
void updateAccList( 
#if NeedFunctionPrototypes
	void 
#endif
);
void deleteAccelerator( 
#if NeedFunctionPrototypes
	int 
#endif
);
void accCmdPopup(
#if NeedFunctionPrototypes
        Widget, XEvent*
#endif
);


extern struct FileManager* fmgr;

static struct Accelerator *accels = (struct Accelerator *)NULL;
static Widget dialog = NULL;
static int isDirty = 0;
static XtTranslations parsed_cmd_xlations = (XtTranslations)NULL;
static char cmdTranslations[] = "#override <Btn3Down>: AccCmdPopupReq()";
static XtActionsRec cmdActions[] = {
                        { "AccCmdPopupReq", (XtActionProc)accCmdPopup }
                        };


void addAccelerator( dInfo )
  DragInfo* dInfo;
{
  struct Accelerator *acc;
  int numm = numAccelerators(accels), i, nFM = numFM(fmgr);
  
  for( i = 0; i < numm; i++ )
    if( !strcmp( getIthAccelerator(accels,i)->di.dragName, dInfo->dragName ) )  return;

  acc = appendAccelerator( &accels );
  isDirty = 1;

  acc->di.dragName = (char *)XtMalloc( strlen(dInfo->dragName)+1 );
  strcpy( acc->di.dragName, dInfo->dragName );
  stat( dInfo->dragName, &acc->s );
  strcpy( acc->di.dirName, dInfo->dirName );
  acc->di.numDragNames = dInfo->numDragNames;
  acc->di.p = dInfo->p;

  for( i = 0; i < nFM; i++ )
    createWidgets( getIthFM(fmgr,i)->topSWRC, acc );

  if( dialog )
    if( XtIsManaged(dialog) )  updateAccList();
}

void createWidgets( w, acc )
  Widget w;
  struct Accelerator *acc;
{
  Widget form, label, nameLabel;
  Pixel fg, bg;
  XmString xms;

  xms = XmStringCreateSimple( acc->di.dirName );
  XtVaGetValues( w, XmNforeground, &fg, XmNbackground, &bg, NULL );

  form = XtVaCreateManagedWidget( "form",
                xmFormWidgetClass, w,
                XmNwidth, 60,
                XmNheight, 85,
                XmNforeground, fg,
                XmNbackground, bg,
		XmNuserData, (XtPointer)acc,
                NULL );

  label = XtVaCreateManagedWidget( "label",
                xmPushButtonWidgetClass, form,
                XmNlabelType, XmPIXMAP,
		XmNlabelPixmap, acc->di.p,
                XmNtopAttachment, XmATTACH_FORM,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 25,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNforeground, fg,
                XmNbackground, bg,
                XmNshadowThickness, 0,
                XmNtraversalOn, False,
		XmNuserData, (XtPointer)acc,
		XmNtranslations, parsed_cmd_xlations,
                NULL );
  registerDragSite( label );
  XtAddCallback( label, XmNactivateCallback, accelCB, NULL );

  if( acc->di.numDragNames == 1 )
  {
    if( S_ISDIR( acc->s.st_mode ) )
      registerDropSite( label, DIR_DROP_SITE );
  }

  nameLabel = XtVaCreateManagedWidget( "nameLabel",
                xmLabelWidgetClass, form,
                XmNlabelString, xms,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, label,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
		XmNforeground, fg,
                XmNbackground, bg,
                NULL );
  unregisterDragSite( nameLabel );
  XmStringFree( xms );
}

void deleteAccelerator( i )
  int i;
{
  if( i < numAccelerators(accels) && i > 0 )
  {
    struct Accelerator *listend = getIthAccelerator( accels, i );
    WidgetList wl;
    int ct, i, numChildren, nFM = numFM(fmgr);

    isDirty = 1;

    for( ct = 0; ct < nFM; ct++ )
    {
      struct FileManager *f = getIthFM( fmgr, ct );
      XtVaGetValues( f->topSWRC, XmNchildren, &wl, 
		XmNnumChildren, &numChildren, NULL );

      for( i = 0; i < numChildren; i++ )
      {
        XtPointer xtp;
	struct Accelerator *acc;
	XtVaGetValues( wl[i], XmNuserData, &xtp, NULL );
	acc = (struct Accelerator *)xtp;
        if( !strcmp( listend->di.dragName, acc->di.dragName ) )
        {
 	  XtDestroyWidget( wl[i] );
	  break;
        }
      }
    }

    if( listend->di.dragName )  XtFree( listend->di.dragName );
    removeIthAccelerator( &accels, i );
  }
}

void deleteAcceleratorByName( name )
  char* name;
{
  int i, num = numAccelerators(accels);

  for( i = 1; i < num; i++ )
  {
    struct Accelerator *acc = getIthAccelerator( accels, i );
    char buffer[256], *ptr = (char *)XtMalloc( strlen(acc->di.dragName)+1 ),
	 *pptr = ptr;
    strcpy( ptr, acc->di.dragName );

    while( sscanf( pptr, "%s", buffer ) == 1 )
    {
      if( !strcmp( name, buffer ) )
      {
        deleteAccelerator( i );
        if( dialog )
        {
	  if( XtIsManaged(dialog) )
	    createDelAccelPopup( (Widget)NULL, (XtPointer)NULL, (XtPointer)NULL );
	  break;
        }
      }

      pptr += strlen( buffer );
      while( pptr[0] == ' ' )  ptr++;
    }

    XtFree( (char *)ptr );
  }
}

void updateAccList( )
{
  XmString *str;
  int i, num = numAccelerators(accels);

  XmListDeleteAllItems( XmSelectionBoxGetChild(dialog,XmDIALOG_LIST) );
  if( num > 1 )
  {
    str = (XmString *)XtMalloc( (num-1) * sizeof(XmString) );
    for( i = 1; i < num; i++ )
      str[i-1] = XmStringCreateSimple( getIthAccelerator(accels,i)->di.dragName );

    XtVaSetValues( dialog,
                 XmNlistItems, str,
                 XmNlistItemCount, num-1,
                 NULL );

    for( i = 1; i < num; i++ )  XmStringFree( str[i-1] );
    XtFree( (char *)str );
  }
}

void createDelAccelPopup( w, xtp1, xtp2 )
  Widget w;
  XtPointer xtp1, xtp2;
{
  XmString *str = (XmString *)NULL;
  int i, num = numAccelerators(accels);

  if( num > 1 )
  {
    str = (XmString *)XtMalloc( (num-1) * sizeof(XmString) );
    for( i = 1; i < num; i++ )
      str[i-1] = XmStringCreateSimple( getIthAccelerator(accels,i)->di.dragName );
  }
  
  if( !dialog )
  {
    XmString t, als, title;
    t = XmStringCreateSimple( "Accelerators" );
    als = XmStringCreateSimple( "Delete" );
    title = XmStringCreateSimple( "Delete Accelerator(s)" );

    dialog = XmCreateSelectionDialog( w, "dialog", NULL, 0 );
    XtUnmanageChild( XmSelectionBoxGetChild(dialog,XmDIALOG_HELP_BUTTON) );
    XtUnmanageChild( XmSelectionBoxGetChild(dialog,XmDIALOG_OK_BUTTON) );
    XtAddCallback( dialog, XmNokCallback, doneCB, NULL );
    XtAddCallback( dialog, XmNapplyCallback, deleteCB, NULL );

    XtVaSetValues( dialog,
		   XmNlistLabelString, t,
		   XmNmustMatch, True,
		   XmNapplyLabelString, als,
		   XmNdialogTitle, title,
		   NULL );

    XmStringFree( t );
    XmStringFree( als );
    XmStringFree( title );
  }

  XmListDeleteAllItems( XmSelectionBoxGetChild(dialog,XmDIALOG_LIST) );
  XmTextFieldSetString( XmSelectionBoxGetChild(dialog,XmDIALOG_TEXT), "" );

  if( num > 1 )
  {
    XtVaSetValues( dialog,
		 XmNlistItems, str,
		 XmNlistItemCount, num-1,
		 NULL );

    for( i = 1; i < num; i++ )  XmStringFree( str[i-1] );
    XtFree( (char *)str );
  }

  if( !XtIsManaged(dialog) )  XtManageChild( dialog );
  XRaiseWindow( XtDisplay(dialog), XtWindow(dialog) );
}

void doneCB( w, xtp1, xtp2 )
  Widget w;
  XtPointer xtp1, xtp2;
{
  XtUnmanageChild( dialog );
}

void deleteCB( w, xtp1, xtp2 )
  Widget w;
  XtPointer xtp1, xtp2;
{
  XmSelectionBoxCallbackStruct *cbs = (XmSelectionBoxCallbackStruct *)xtp2;
  char *value;
  int i, num = numAccelerators(accels);

  if( num < 2 )  return;

  XmStringGetLtoR( cbs->value, XmSTRING_DEFAULT_CHARSET, &value );
  for( i = 1; i < num; i++ )
  {
    struct Accelerator *acc = getIthAccelerator( accels, i );
    if( !strcmp( value, acc->di.dragName ) )
    {
      deleteAccelerator( i );
      createDelAccelPopup( (Widget)NULL, (XtPointer)NULL, (XtPointer)NULL );
      break;
    }
  }

  XtFree( value );
}

void accelCB( w, xtp1, xtp2 )
  Widget w;
  XtPointer xtp1, xtp2;
{
  XmPushButtonCallbackStruct *cbs = (XmPushButtonCallbackStruct *)xtp2;
  static XtIntervalId id;
  XtAppContext appContext = XtWidgetToApplicationContext( w );

  if( cbs->click_count == 2 )
  {
    struct Accelerator *acc;
    struct FileManager *fm;
    XtPointer xt;

    XtRemoveTimeOut( id );

    XtVaGetValues( w, XmNuserData, &xt, NULL );
    acc = (struct Accelerator *)xt;
    XtVaGetValues( XtParent(XtParent(w)), XmNuserData, &xt, NULL );
    fm = (struct FileManager *)xt;

    if( acc->di.numDragNames == 1 )  
    {
      struct FileManager *fm;
      XtPointer xt;
      int rix = 0;

      setWatchCursor();
      XtVaGetValues( XtParent(XtParent(w)), XmNuserData, &xt, NULL );
      fm = (struct FileManager *)xt;
      if( fm->runInXterm )  rix = 1;
      forkNexec( w, &acc->s, acc->di.dragName, rix );
      unsetWatchCursor();
    }
  }
  else if( cbs->click_count == 1 )
  {
    id = XtAppAddTimeOut( appContext, 1000, (XtTimerCallbackProc)singleClick,
		(XtPointer)w );
  }
}

void singleClick( xtp )
  XtPointer xtp;
{
  struct Accelerator *acc;
  struct FileManager *fm;
  XtPointer xt;
  Widget w = (Widget)xtp;

  setWatchCursor();
  XtVaGetValues( w, XmNuserData, &xt, NULL );
  acc = (struct Accelerator *)xt;
  XtVaGetValues( XtParent(XtParent(w)), XmNuserData, &xt, NULL );
  fm = (struct FileManager *)xt;
  accelSelect( &acc->di, fm );
  unsetWatchCursor();
}

void writeRCFile( )
{
  FILE *fd;
  char dirBuffer[256];
  int i, n = numAccelerators(accels);

  sprintf( dirBuffer, "%s/%s", getenv("HOME"), RC_FILE );
  if( (fd = fopen(dirBuffer,"w")) == NULL )
  {
    fprintf( stderr, "***MFM ERROR: Cannot open %s\n", dirBuffer );
    return;
  }

  for( i = 0; i < n; i++ )
  {
      struct Accelerator *listend = getIthAccelerator( accels, i );
      fprintf( fd, "NUM: %d\nLABEL: %s\nFILES: %s\n", listend->di.numDragNames, 
		listend->di.dirName, listend->di.dragName );
  }

  isDirty = 0;
  fclose( fd );
}

void loadAccels( w )
  Widget w;
{
  int i, num = numAccelerators(accels);

  for( i = 0; i < num; i++ )
  {
    struct Accelerator *acc = getIthAccelerator( accels, i );
    createWidgets( w, acc );
  }
}

void readRCFile( w ) 
  Widget w;
{
  FILE *fd;
  char dirBuffer[1000], tmp[10];
  struct stat s;
  DragInfo di;

  isDirty = 0;

  parsed_cmd_xlations = XtParseTranslationTable( cmdTranslations );
  XtAppAddActions( XtWidgetToApplicationContext(w), cmdActions,
                        XtNumber(cmdActions) );

  di.dragName = (char *)XtMalloc( 1000 );

  /*  add home directory as first accelerator  */
  strcpy( di.dragName, getenv("HOME") );
  strcpy( di.dirName, strrchr( di.dragName, '/' ) + 1 );
  di.numDragNames = 1;
  di.p = getPixmap( "HOME", s, LOCAL_FILESYSTEM, "" );
  addAccelerator( &di );

  sprintf( dirBuffer, "%s/%s", getenv("HOME"), RC_FILE );
  if( (fd = fopen(dirBuffer,"r")) == NULL )  return;

  while( fgets( dirBuffer, sizeof(dirBuffer), fd ) != NULL )
  {
    if( strcmp( dirBuffer, "" ) && strcmp( dirBuffer, "\n" ) &&
        sscanf( dirBuffer, "%s", tmp ) == 1 )
    {
      int i = 0;
      while( dirBuffer[i] != '\0' )
      {
        if( dirBuffer[i] == '\n' )  dirBuffer[i] = '\0';
        i++;
      }

      if( !strcmp( tmp, "NUM:" ) )
        sscanf( dirBuffer, "%*s %d", &di.numDragNames );
      else if( !strcmp( tmp, "LABEL:" ) )
        strcpy( di.dirName, dirBuffer+7 );
      else if( !strcmp( tmp, "FILES:" ) )
      {
        strcpy( di.dragName, dirBuffer+7 );

        if( di.numDragNames > 1 )
	{
	  di.p = getPixmap( "MULTIPLE", s, LOCAL_FILESYSTEM, "" );
          addAccelerator( &di );
	}
        else
        {
          if( stat( di.dragName, &s ) < 0 )
            fprintf( stderr, "Cannot stat %s\n", di.dragName );
          else
	  {
	    di.p = getPixmap( di.dragName, s, LOCAL_FILESYSTEM, "" );
            addAccelerator( &di );
	  }
        }
      }
    }
  }

  fclose( fd );
  XtFree( (char *)di.dragName );
}

void accCmdPopup( w, e )
  Widget w;
  XEvent* e;
{
  XtPointer xtp;
  struct Accelerator *acc;
  char cmd[1056];

  XtVaGetValues( w, XmNuserData, &xtp, NULL );
  acc = (struct Accelerator *)xtp;
  memset( (void *)cmd, '\0', sizeof(cmd) );
  getCommand( acc->di.dragName, cmd );
  if( strcmp( cmd, "" ) )  popupCmdWithCmd( w, cmd );
}
