#include "globals.h"
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <Xm/AtomMgr.h>

void startDrag( 
#if NeedFunctionPrototypes
	Widget, XEvent * 
#endif
);
void noDrag( 
#if NeedFunctionPrototypes
	Widget, XEvent * 
#endif
);
void TransferProc( 
#if NeedFunctionPrototypes
	Widget, XtPointer, Atom *, Atom *, XtPointer, unsigned long *, int 
#endif
);
void HandleDrop( 
#if NeedFunctionPrototypes
	Widget, XtPointer, XtPointer 
#endif
);
Boolean DragConvertProc( 
#if NeedFunctionPrototypes
	Widget, Atom *, Atom *, Atom *, XtPointer *,
	unsigned long *, int *, unsigned long *, XtPointer, 
	XtRequestId * 
#endif
);

static Atom ATOM_STRING, ATOM_PIXMAP, ATOM_INTEGER, ATOM_LABEL;
static Atom ATOM_FTP_STRING, ATOM_FTP_INTEGER, ATOM_FTP_LABEL;
static int currentOperation, currentType;
static XtTranslations drag_xlations, noDrag_xlations;
static DragInfo dragInfo;

static char dragTranslations[] = "#override <Btn2Down>: StartDrag()";
static char noDragTranslations[] = "#override <Btn2Down>: NoDrag()";
static XtActionsRec dragActions[] = {
                  { "StartDrag", (XtActionProc)startDrag }
                  };
static XtActionsRec noDragActions[] = {
		  { "NoDrag", (XtActionProc)noDrag }
		  };


void initDD( toplevel )
  Widget toplevel;
{
  XtAppContext appContext = XtWidgetToApplicationContext( toplevel );
  drag_xlations = XtParseTranslationTable( dragTranslations );
  noDrag_xlations = XtParseTranslationTable( noDragTranslations );
  XtAppAddActions( appContext, dragActions, XtNumber( dragActions ) );
  XtAppAddActions( appContext, noDragActions, XtNumber( noDragActions ) );
  ATOM_STRING = XmInternAtom( XtDisplay(toplevel), "STRING", False );
  ATOM_FTP_STRING = XmInternAtom( XtDisplay(toplevel), "FTP_STRING", False );
  ATOM_PIXMAP = XmInternAtom( XtDisplay(toplevel), "PIXMAP", False );
  ATOM_INTEGER = XmInternAtom( XtDisplay(toplevel), "INTEGER", False );
  ATOM_FTP_INTEGER = XmInternAtom( XtDisplay(toplevel), "FTP_INTEGER", False );
  ATOM_LABEL = XmInternAtom( XtDisplay(toplevel), "COMPOUND_TEXT", False );
  ATOM_FTP_LABEL = XmInternAtom( XtDisplay(toplevel), "COMPOUND_TEXT", False );
}

void registerDragSite( w )
  Widget w;
{
  XtOverrideTranslations( w, drag_xlations );
}

void unregisterDragSite( w )
  Widget w;
{
  XtOverrideTranslations( w, noDrag_xlations );
}

void noDrag( w, event )
  Widget w;
  XEvent* event;
{
}

void startDrag( w, event )
  Widget w;
  XEvent* event;
{
  Cardinal n;
  Atom exportList[4];
  Widget drag_widget, pp;
  unsigned int pixmap_width, pixmap_height, pixmap_depth, bdw;
  Window win;
  int x, y, i;
  Arg args[10];
  XtPointer xtp;
  struct FileManager *f;

  XtVaGetValues( XtParent(XtParent(w)), XmNuserData, &xtp, NULL );
  f = (struct FileManager *)xtp;
  XtVaGetValues( w, XmNuserData, &xtp, NULL );
  pp = XtParent( XtParent(w) );

  if( pp == f->topSWRC )
  {
    struct Accelerator *acc = (struct Accelerator *)xtp;
    dragInfo.p = acc->di.p;
    dragInfo.numDragNames = acc->di.numDragNames;
    strcpy( dragInfo.dirName, acc->di.dirName );
    if( dragInfo.dragName )  XtFree( (char *)dragInfo.dragName );
    dragInfo.dragName = (char *)XtMalloc( strlen(acc->di.dragName)+1 );
    strcpy( dragInfo.dragName, acc->di.dragName );
  }
  else if( pp == f->form )
  {
    struct DirInfo *dInfo = (struct DirInfo *)xtp;
    dragInfo.p = dInfo->di.p;
    dragInfo.numDragNames = dInfo->di.numDragNames;
    strcpy( dragInfo.dirName, dInfo->di.dirName );
    if( dragInfo.dragName )  XtFree( (char *)dragInfo.dragName );
    dragInfo.dragName = (char *)XtMalloc( strlen(dInfo->di.dragName)+1 );
    strcpy( dragInfo.dragName, dInfo->di.dragName );
  }
  else
    return;

  currentType = f->type;
  if( f->type == FTP_FILESYSTEM )  strcpy( dragInfo.dirName, f->host );

  XGetGeometry( XtDisplay(w), dragInfo.p, &win, &x, &y,
        &pixmap_width, &pixmap_height, &bdw, &pixmap_depth );

  n = 0;
  XtSetArg( args[n], XmNpixmap, dragInfo.p );  n++;
  XtSetArg( args[n], XmNwidth, pixmap_width );  n++;
  XtSetArg( args[n], XmNheight, pixmap_height );  n++;
  XtSetArg( args[n], XmNdepth, pixmap_depth );  n++;
  drag_widget = XmCreateDragIcon( w, "drag_widget", args, n );

  n = 0;
  XtSetArg( args[n], XmNblendModel, XmBLEND_ALL );  n++;
  XtSetArg( args[n], XmNsourcePixmapIcon, drag_widget );  n++;
  XtSetArg( args[n], XmNconvertProc, DragConvertProc );  n++;
  if( f->type == LOCAL_FILESYSTEM )
  {
    exportList[0] = ATOM_LABEL;
    exportList[1] = ATOM_INTEGER;
    exportList[2] = ATOM_PIXMAP;
    exportList[3] = ATOM_STRING;
    XtSetArg( args[n], XmNexportTargets, exportList );  n++;
    XtSetArg( args[n], XmNnumExportTargets, 4 );  n++;
    XtSetArg( args[n], XmNdragOperations, XmDROP_COPY|XmDROP_MOVE|XmDROP_LINK );  n++;
  }
  else
  {
    exportList[0] = ATOM_FTP_LABEL;
    exportList[1] = ATOM_FTP_INTEGER;
    exportList[2] = ATOM_FTP_STRING;
    XtSetArg( args[n], XmNexportTargets, exportList );  n++;
    XtSetArg( args[n], XmNnumExportTargets, 3 );  n++;
    XtSetArg( args[n], XmNdragOperations, XmDROP_COPY );  n++;
  }
  XmDragStart( w, event, args, n );
}

Boolean DragConvertProc( w, selection, target, typeRtn, valueRtn,
                lengthRtn, formatRtn, max_lengthRtn, client_data, request_id )
  Widget w;
  Atom *selection, *target, *typeRtn;
  XtPointer *valueRtn, client_data;
  unsigned long *lengthRtn, *max_lengthRtn;
  int *formatRtn;
  XtRequestId *request_id;
{
  *formatRtn = 32;
  *typeRtn = *target;

  if( ( *target == ATOM_STRING && currentType == LOCAL_FILESYSTEM ) || 
      ( *target == ATOM_FTP_STRING && currentType == FTP_FILESYSTEM ) )
  {
    *valueRtn = (XtPointer)(dragInfo.dragName);
    *lengthRtn = strlen(dragInfo.dragName)+1;
    return( True );
  }
  else if( ( *target == ATOM_INTEGER && currentType == LOCAL_FILESYSTEM ) || 
           ( *target == ATOM_FTP_INTEGER && currentType == FTP_FILESYSTEM ) )
  {
    *valueRtn = (XtPointer)(&dragInfo.numDragNames);
    *lengthRtn = sizeof( &dragInfo.numDragNames );
    return( True );
  }
  else if( *target == ATOM_PIXMAP )
  {
    *valueRtn = (XtPointer)(&dragInfo.p);
    *lengthRtn = sizeof( &dragInfo.p );
    return( True );
  }
  else if( *target == ATOM_LABEL || *target == ATOM_FTP_LABEL )
  {
    *valueRtn = (XtPointer)dragInfo.dirName;
    *lengthRtn = sizeof( dragInfo.dirName );
    return( True );
  }
  else
    return( False );
}

void registerDropSite( w, b )
  Widget w;
  int b;
{
  Atom importList[5];
  Arg args[6];
  Cardinal n = 0;

  if( b == DEV_DROP_SITE )
  {
    importList[0] = ATOM_INTEGER;
    importList[1] = ATOM_STRING;

    XtSetArg( args[0], XmNdropSiteOperations, XmDROP_LINK );
    XtSetArg( args[1], XmNnumImportTargets, 2 );
    n += 2;
  }
  else if( b == ACC_DROP_SITE )
  {
    importList[0] = ATOM_LABEL;
    importList[1] = ATOM_INTEGER;
    importList[2] = ATOM_PIXMAP;
    importList[3] = ATOM_STRING;

    XtSetArg( args[0], XmNdropSiteOperations, XmDROP_LINK );
    XtSetArg( args[1], XmNnumImportTargets, 4 );
    XtSetArg( args[2], XmNanimationStyle, XmDRAG_UNDER_SHADOW_IN );
    n += 3;
  }
  else if( b == DIR_DROP_SITE )
  {
    importList[0] = ATOM_INTEGER;
    importList[1] = ATOM_STRING;
    importList[2] = ATOM_FTP_LABEL;
    importList[3] = ATOM_FTP_INTEGER;
    importList[4] = ATOM_FTP_STRING;

    XtSetArg( args[0], XmNdropSiteOperations, XmDROP_LINK|XmDROP_COPY|XmDROP_MOVE );
    XtSetArg( args[1], XmNnumImportTargets, 5 );
    n += 2;
  }

  XtSetArg( args[n], XmNimportTargets, importList );  n++;
  XtSetArg( args[n], XmNdropProc, HandleDrop );  n++;
  XmDropSiteRegister( w, args, n );
}

void TransferProc( w, closure, seltype, type, value, length, format )
  Widget w;
  XtPointer closure, value;
  Atom *seltype, *type;
  unsigned long *length;
  int format;
{
  static DragInfo di;

  if( *type == ATOM_STRING || *type == ATOM_FTP_STRING )
  {
    Widget dw = (Widget)closure;
    di.dragName = (char *)XtMalloc( strlen((char *)value) + 1 );
    strcpy( di.dragName, (char *)value );

    if( isAddAccWidget(dw) )
    {
      if( di.numDragNames > 0 && strcmp( di.dirName, "" ) && di.p )
      {
        addAccelerator( &di );
	writeRCFile();
      }
      else
	fprintf( stderr, "mfm: Cannot add accelerator with NULL values...\n" );
    }
    else if( isaDeviceWidget(dw) )
    {
      struct Device *d;
      XtPointer xtp;
      XtVaGetValues( dw, XmNuserData, &xtp, NULL );
      d = (struct Device *)xtp;
      execDeviceCommand( dw, d, di.numDragNames, di.dragName );
    }
    else if( *type == ATOM_STRING )
    {
      struct FileManager *f;
      XtPointer xtp;
      DragInfo *ddi;
      char *cmd, *ptr = (char *)value, buffer[256];

      XtVaGetValues( XtParent(XtParent(dw)), XmNuserData, &xtp, NULL );
      f = (struct FileManager *)xtp;
      XtVaGetValues( dw, XmNuserData, &xtp, NULL );

      if( XtParent(XtParent(dw)) == f->topSWRC )
      {
        struct Accelerator *acc = (struct Accelerator *)xtp;
        ddi = &acc->di;
      }
      else if( XtParent(XtParent(dw)) == f->form )
      {
        struct DirInfo *dInfo = (struct DirInfo *)xtp;
        ddi = &dInfo->di;
      }
      else
      {
	di.numDragNames = -1;
	di.p = (Pixmap)NULL;
	strcpy( di.dirName, "" );
	XtFree( di.dragName );
        return;
      }

      while( sscanf( ptr, "%s", buffer ) == 1 )
      {
        if( currentOperation == XmDROP_COPY )
        {
	  cmd = (char *)XtMalloc( strlen(buffer) +
		strlen(ddi->dragName) + 25 );
	  sprintf( cmd, "cp -r %s %s 2>&1", buffer, ddi->dragName );
        }
        else if( currentOperation == XmDROP_MOVE )
        {
	  cmd = (char *)XtMalloc( strlen(buffer) +
                strlen(ddi->dragName) + 25 );
          sprintf( cmd, "mv -f %s %s 2>&1", buffer, ddi->dragName );
        }
        else if( currentOperation == XmDROP_LINK )
        {
	  char *name = strrchr( buffer, '/' );
	  cmd = (char *)XtMalloc( strlen(buffer) +
                strlen(ddi->dragName) + strlen(name) + 25 );
	  sprintf( cmd, "ln -s %s %s/%s 2>&1", buffer, ddi->dragName, name );
	  XtFree( (char *)name );
        }

        if( issuePipeCommand( cmd ) )
	{
	  XtFree( (char *)cmd );
	  return;
	}

        XtFree( (char *)cmd );
        ptr += strlen( buffer );
        while( ptr[0] == ' ' )  ptr++;
      }
    }
    else if( *type == ATOM_FTP_STRING )
    {
      struct FileManager *f;
      XtPointer xtp;
      struct DirInfo *dInfo;
      char *cmd, buffer[256];
      FILE *fd;

      cmd = (char *)XtMalloc( strlen(di.dragName)+10 );
      sprintf( cmd, "RETR %s", di.dragName );

      XtVaGetValues( XtParent(XtParent(dw)), XmNuserData, &xtp, NULL );
      f = (struct FileManager *)xtp;
      XtVaGetValues( dw, XmNuserData, &xtp, NULL );
      dInfo = (struct DirInfo *)xtp;

      sprintf( buffer, "%s/%s", dInfo->di.dragName, strrchr(di.dragName,'/') );
      if( (fd = fopen(buffer,"w")) == (FILE *)NULL )
      {
        char errorString[300];
        sprintf( errorString, "Cannot Open %s\nTransfer Aborted...", buffer );
	popupErrorDialog( errorString );
      }
      else
      {
        fclose( fd );
        if( access( buffer, F_OK ) == 0 )  remove( buffer );
        ftpSendCommand( getFileMgrByFtpHost(di.dirName), cmd, -1, buffer );
      }

      XtFree( (char *)cmd );
    }

    updateAllFileManagers();

    di.numDragNames = -1;
    di.p = (Pixmap)NULL;
    strcpy( di.dirName, "" );
    XtFree( di.dragName );
  }
  else if( *type == ATOM_INTEGER || *type == ATOM_FTP_INTEGER )
  {
    di.numDragNames = *((int *)value);
  }
  else if( *type == ATOM_PIXMAP )
  {
    di.p = *((Pixmap *)value);
  }
  else if( *type == ATOM_LABEL || *type == ATOM_FTP_LABEL )
  {
    strcpy( di.dirName, (char *)value );
  }
}

void HandleDrop( w, client_data, call_data )
  Widget w;
  XtPointer client_data, call_data;
{
  XmDropProcCallback DropData;
  XmDropTransferEntryRec transferEntries[5];
  XmDropTransferEntry transferList;
  Cardinal n = 0;
  Arg args[5];

  DropData = ( XmDropProcCallback )call_data;

  if( DropData->dropAction != XmDROP )
  {
    XtSetArg( args[n], XmNtransferStatus, XmTRANSFER_FAILURE );  n++;
  }
  else if( isAddAccWidget(w) && DropData->operation != XmDROP_LINK )
  {
    XtSetArg( args[n], XmNtransferStatus, XmTRANSFER_FAILURE );  n++;
  }
  else if( DropData->operation == XmDROP_NOOP )
  {
    XtSetArg( args[n], XmNtransferStatus, XmTRANSFER_FAILURE );  n++;
  }
  else 
  {
    currentOperation = DropData->operation;

    if( isAddAccWidget(w) )
    {
      transferEntries[0].target = ATOM_LABEL;
      transferEntries[0].client_data = (XtPointer) w;
      transferEntries[1].target = ATOM_INTEGER;
      transferEntries[1].client_data = (XtPointer) w;
      transferEntries[2].target = ATOM_PIXMAP;
      transferEntries[2].client_data = (XtPointer) w;
      transferEntries[3].target = ATOM_STRING;
      transferEntries[3].client_data = (XtPointer) w;
      transferList = transferEntries;
      XtSetArg( args[n], XmNnumDropTransfers, 4 );  n++;
    }
    else if( isaDeviceWidget(w) )
    {
      transferEntries[0].target = ATOM_INTEGER;
      transferEntries[0].client_data = (XtPointer) w;
      transferEntries[1].target = ATOM_STRING;
      transferEntries[1].client_data = (XtPointer) w;
      transferList = transferEntries;
      XtSetArg( args[n], XmNnumDropTransfers, 2 );  n++;
    }
    else
    {
      transferEntries[0].target = ATOM_INTEGER;
      transferEntries[0].client_data = (XtPointer) w;
      transferEntries[1].target = ATOM_STRING;
      transferEntries[1].client_data = (XtPointer) w;
      transferEntries[2].target = ATOM_FTP_LABEL;
      transferEntries[2].client_data = (XtPointer) w;
      transferEntries[3].target = ATOM_FTP_INTEGER;
      transferEntries[3].client_data = (XtPointer) w;
      transferEntries[4].target = ATOM_FTP_STRING;
      transferEntries[4].client_data = (XtPointer) w;
      transferList = transferEntries;
      XtSetArg( args[n], XmNnumDropTransfers, 5 );  n++;
    }

    XtSetArg( args[n], XmNdropTransfers, transferList );  n++;
    XtSetArg( args[n], XmNtransferProc, TransferProc );  n++;
  }

  XmDropTransferStart( DropData->dragContext, args, n );
}
